import { DecoderError, DecoderResult, DecoderSuccess } from './DecoderResult';
export declare type DecodeFnResult<T> =
  | DecoderSuccess<T>
  | DecoderError
  | DecoderError[];
/**
 * An object which can be used to validate and process `unknown` values
 * and cast them to the appropriate typescript type.
 *
 * @param decodeFn the function which is used to decode input values
 */
export declare class Decoder<R, I = any> {
  readonly decodeFn: (input: I) => DecodeFnResult<R>;
  constructor(decodeFn: (input: I) => DecodeFnResult<R>);
  decode(input: Promise<I>): Promise<DecoderResult<R>>;
  decode(input: I): DecoderResult<R>;
  /**
   * On decode failure, handle the DecoderErrors.
   */
  catch<K>(
    fn: (input: unknown, errors: DecoderError[]) => DecodeFnResult<K>,
  ): Decoder<R | K, I>;
  /**
   * On decode success, transform a value using the provided
   * transformation function.
   */
  map<K>(fn: (input: R) => K): Decoder<K, I>;
  /**
   * On decode success, perform a new validation check.
   */
  chain<K>(fn: (input: R) => DecodeFnResult<K>): Decoder<K, I>;
  chain<K>(decoder: Decoder<K, R>): Decoder<K, I>;
  toAsyncDecoder(): AsyncDecoder<R, I>;
}
/**
 * An object which can be used to validate and process `unknown` values
 * and cast them to the appropriate typescript type. Unlike `Decoder`,
 * `AsyncDecoder` can receive a `decodeFn` which returns a promise.
 *
 * @param decodeFn the function which is used to decode input values
 */
export declare class AsyncDecoder<R, I = any> {
  readonly decodeFn: (input: I) => Promise<DecodeFnResult<R>>;
  constructor(decodeFn: (input: I) => Promise<DecodeFnResult<R>>);
  decode(input: I | Promise<I>): Promise<DecoderResult<R>>;
  /**
   * On decode failure, handle the DecoderErrors.
   */
  catch<K>(
    fn: (
      input: unknown,
      errors: DecoderError[],
    ) => DecodeFnResult<K> | Promise<DecodeFnResult<K>>,
  ): AsyncDecoder<R | K, I>;
  /**
   * On decode success, transform a value using the provided
   * transformation function.
   */
  map<K>(fn: (input: R) => K | Promise<K>): AsyncDecoder<K, I>;
  /**
   * On decode success, perform a new validation check.
   */
  chain<K>(
    fn: (input: R) => DecodeFnResult<K> | Promise<DecodeFnResult<K>>,
  ): AsyncDecoder<K, I>;
  chain<K>(decoder: Decoder<K, R> | AsyncDecoder<K, R>): AsyncDecoder<K, I>;
}
export declare type DecoderReturnType<T> = T extends Decoder<infer R>
  ? R
  : T extends AsyncDecoder<infer R>
  ? R
  : unknown;
export declare type DecoderInputType<T> = T extends Decoder<any, infer I>
  ? I
  : T extends AsyncDecoder<any, infer I>
  ? I
  : unknown;
// # sourceMappingURL=Decoder.d.ts.map
