import {
  DecoderError,
  areDecoderErrors,
  AsyncDecoder,
  DecoderSuccess,
  Decoder,
  isDecoderSuccess,
} from 'ts-decoders';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
  return new (P || (P = Promise))(function(resolve, reject) {
    function fulfilled(value) {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    }
    function rejected(value) {
      try {
        step(generator['throw'](value));
      } catch (e) {
        reject(e);
      }
    }
    function step(result) {
      result.done
        ? resolve(result.value)
        : new P(function(resolve) {
            resolve(result.value);
          }).then(fulfilled, rejected);
    }
    step((generator = generator.apply(thisArg, _arguments || [])).next());
  });
}

function applyOptionsToDecoderErrors(input, errors, options) {
  if (errors instanceof DecoderError) {
    errors = [errors];
  }
  if (!options) return errors;
  if (options.decoderName !== undefined && options.data !== undefined) {
    Array.from(errors.values()).forEach(error => {
      error.data = options.data;
      error.decoderName = options.decoderName;
    });
  } else if (options.decoderName) {
    Array.from(errors.values()).forEach(error => {
      error.decoderName = options.decoderName;
    });
  } else if (options.data !== undefined) {
    Array.from(errors.values()).forEach(error => {
      error.data = options.data;
    });
  }
  if (typeof options.errorMsg === 'function') {
    errors = options.errorMsg(input, errors);
    if (errors instanceof DecoderError) errors = [errors];
  } else if (options.errorMsg) {
    Array.from(errors.values()).forEach(error => {
      error.message = options.errorMsg;
      return error;
    });
  }
  if (errors.length === 0) {
    throw new Error(
      'Provided DecoderError function must return ' +
        'an array of DecoderErrors with length greater than 0.',
    );
  }
  return errors;
}
function _raceToDecoderSuccess(promises, errors = []) {
  return __awaiter(this, void 0, void 0, function*() {
    if (promises.length === 0) return errors;
    const indexedPromises = promises.map((promise, index) =>
      promise.then(value => [index, value]),
    );
    const res = yield Promise.race(indexedPromises);
    if (areDecoderErrors(res[1])) {
      promises.splice(res[0], 1);
      errors.push(...res[1]);
      return _raceToDecoderSuccess(promises, errors);
    }
    return res[1];
  });
}
/**
 * Resolves the provided DecoderResults, returning the first
 * DecoderSuccess or all the DecoderErrors.
 */
function raceToDecoderSuccess(promises) {
  return _raceToDecoderSuccess(promises);
}

const decoderName = 'anyOfD';
function childErrors(input, children) {
  return children.map(
    (child, index) =>
      new DecoderError(input, `anyOf ${index}`, 'invalid value', {
        child,
      }),
  );
}
function anyOfD(decoders, provOptions = {}) {
  const options = Object.assign({}, provOptions, {
    decoderName: provOptions.decoderName || decoderName,
    allErrors: true,
  });
  const hasAsyncDecoder = decoders.some(
    decoder => decoder instanceof AsyncDecoder,
  );
  if (hasAsyncDecoder) {
    if (options.decodeInParallel) {
      return new AsyncDecoder(input =>
        __awaiter(this, void 0, void 0, function*() {
          const promises = decoders.map(decoder =>
            __awaiter(this, void 0, void 0, function*() {
              return yield decoder.decode(input);
            }),
          );
          const result = yield raceToDecoderSuccess(promises);
          if (result instanceof DecoderSuccess) return result;
          return applyOptionsToDecoderErrors(
            input,
            childErrors(input, result),
            options,
          );
        }),
      );
    }
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        const errors = [];
        for (const decoder of decoders) {
          const result = yield decoder.decode(input);
          if (result instanceof DecoderSuccess) return result;
          errors.push(...result);
        }
        return applyOptionsToDecoderErrors(
          input,
          childErrors(input, errors),
          options,
        );
      }),
    );
  }
  return new Decoder(input => {
    const errors = [];
    for (const decoder of decoders) {
      const result = decoder.decode(input);
      if (result instanceof DecoderSuccess) return result;
      errors.push(...result);
    }
    return applyOptionsToDecoderErrors(
      input,
      childErrors(input, errors),
      options,
    );
  });
}

/** Convenience function for building a DecoderSuccess result */
function ok(value) {
  return new DecoderSuccess(value);
}
/** Convenience function for building a simple DecoderError[] result */
function err(input, type, errorMsg, decoderName, options) {
  return applyOptionsToDecoderErrors(
    input,
    new DecoderError(input, type, errorMsg, { decoderName }),
    options,
  );
}
/**
 * Builds the `DecoderError#location` property given the child error's
 * location and the current key.
 */
function errorLocation(key, childLocation) {
  // simple check to see if we can render the key using dot (`.`) notation
  const keyIsValidDotAccessor =
    typeof key === 'string' && /^[a-zA-Z]+$/.test(key);
  // prettier-ignore : prettier makes this harder to read
  const location = keyIsValidDotAccessor
    ? key.toString()
    : // if we can't render the key using dot notation, is it a string?
    typeof key === 'string'
    ? `["${key}"]`
    : `[${key}]`;
  // check to see if the previous nested child location was rendered using dot
  // notation. A '[' as the first character indicates that it was not.
  // `undefined` as the first character means the child location === ''.
  // If it was, add the dot (`.`) in.
  return ['[', undefined].includes(childLocation[0])
    ? location.concat(childLocation)
    : `${location}.${childLocation}`;
}

/**
 * Creates a decoder which always returns `DecoderSuccess` with
 * whatever input value is provided to it.
 */
function anyD() {
  return new Decoder(value => ok(value));
}

const decoderName$1 = 'arrayD';
function nonArrayError(input, options = {}) {
  return applyOptionsToDecoderErrors(
    input,
    new DecoderError(input, 'invalid type', 'must be an array'),
    options,
  );
}
function childError(input, child, key) {
  const location = errorLocation(key, child.location);
  return new DecoderError(
    input,
    'invalid element',
    `invalid element [${key}] > ${child.message}`,
    {
      child,
      location,
      key,
    },
  );
}
function arrayD(decoder, options = {}) {
  options = Object.assign({}, options, {
    decoderName: options.decoderName || decoderName$1,
  });
  if (!(decoder instanceof Decoder || decoder instanceof AsyncDecoder)) {
    return new Decoder(input =>
      Array.isArray(input) ? ok(input.slice()) : nonArrayError(input, options),
    );
  }
  if (decoder instanceof AsyncDecoder) {
    if (options.allErrors) {
      return new AsyncDecoder(input =>
        __awaiter(this, void 0, void 0, function*() {
          if (!Array.isArray(input)) return nonArrayError(input, options);
          let hasError = false;
          const results = yield Promise.all(
            input.map(item =>
              __awaiter(this, void 0, void 0, function*() {
                const result = yield decoder.decode(item);
                if (!hasError && areDecoderErrors(result)) {
                  hasError = true;
                }
                return result;
              }),
            ),
          );
          if (hasError) {
            const errors = [];
            results.forEach((result, index) => {
              if (areDecoderErrors(result)) {
                errors.push(
                  ...result.map(error => childError(input, error, index)),
                );
              }
            });
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          const elements = results.map(result => result.value);
          return ok(elements);
        }),
      );
    }
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        if (!Array.isArray(input)) return nonArrayError(input, options);
        const elements = [];
        let index = -1;
        for (const el of input) {
          index++;
          const result = yield decoder.decode(el);
          if (areDecoderErrors(result)) {
            const errors = result.map(error => childError(input, error, index));
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          elements.push(result.value);
        }
        return ok(elements);
      }),
    );
  }
  return new Decoder(input => {
    if (!Array.isArray(input)) return nonArrayError(input, options);
    const elements = [];
    let index = -1;
    const allErrors = [];
    for (const el of input) {
      index++;
      const result = decoder.decode(el);
      if (areDecoderErrors(result)) {
        const errors = result.map(error => childError(input, error, index));
        if (!options.allErrors) {
          return applyOptionsToDecoderErrors(input, errors, options);
        }
        allErrors.push(...errors);
        continue;
      }
      elements.push(result.value);
    }
    if (allErrors.length > 0) {
      return applyOptionsToDecoderErrors(input, allErrors, options);
    }
    return ok(elements);
  });
}

/** Can be used to verify that an unknown value is a `boolean`. */
function booleanD(options = {}) {
  return new Decoder(value =>
    typeof value === 'boolean'
      ? ok(value)
      : err(value, 'invalid type', 'must be a boolean', 'booleanD', options),
  );
}

const decoderName$2 = 'chainOfD';
function fromChildErrors(input, children) {
  return children.map(
    child =>
      new DecoderError(input, child.type, child.message, {
        decoderName: decoderName$2,
        child,
      }),
  );
}
function chainOfD(a, b, c, d, e, f, o) {
  let options;
  const decoders = [];
  if (Array.isArray(a)) decoders.push(...a);
  else decoders.push(a);
  [b, c, d, e, f, o].forEach(val => {
    if (isDecoder(val)) decoders.push(val);
    else if (val) options = val;
  });
  options = options || {};
  options = Object.assign({}, options, {
    decoderName: options.decoderName || decoderName$2,
  });
  if (decoders.some(decoder => decoder instanceof AsyncDecoder)) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        let result = ok(input);
        for (const decoder of decoders) {
          if (!(result instanceof DecoderSuccess)) break;
          result = yield decoder.decode(result.value);
        }
        if (areDecoderErrors(result)) {
          return applyOptionsToDecoderErrors(
            input,
            fromChildErrors(input, result),
            options,
          );
        }
        return result;
      }),
    );
  }
  return new Decoder(input => {
    const result = decoders.reduce((prev, curr) => {
      return prev instanceof DecoderSuccess ? curr.decode(prev.value) : prev;
    }, ok(input));
    if (areDecoderErrors(result)) {
      return applyOptionsToDecoderErrors(
        input,
        fromChildErrors(input, result),
        options,
      );
    }
    return result;
  });
}
function isDecoder(a) {
  return a instanceof Decoder || a instanceof AsyncDecoder;
}

function constantD(constant) {
  return new Decoder(() => ok(constant));
}

const decoderName$3 = 'dictionaryD';
function childKeyError(input, child, key) {
  const location = errorLocation(key, '');
  return new DecoderError(
    input,
    'invalid key',
    `invalid key ["${key}"] > ${child.message}`,
    {
      decoderName: decoderName$3,
      child,
      location,
      key,
    },
  );
}
function childValueError(input, child, key) {
  const location = errorLocation(key, child.location);
  return new DecoderError(
    input,
    'invalid key value',
    `invalid value for key ["${key}"] > ${child.message}`,
    {
      decoderName: decoderName$3,
      child,
      location,
      key,
    },
  );
}
function asyncDecodeKey(input, decoder, key) {
  return __awaiter(this, void 0, void 0, function*() {
    const keyResult = yield decoder.decode(key);
    if (areDecoderErrors(keyResult)) {
      return keyResult.map(error => childKeyError(input, error, key));
    }
    return keyResult;
  });
}
function asyncDecodeValue(input, decoder, value, key) {
  return __awaiter(this, void 0, void 0, function*() {
    const valueResult = yield decoder.decode(value);
    if (areDecoderErrors(valueResult)) {
      return valueResult.map(error => childValueError(input, error, key));
    }
    return valueResult;
  });
}
function nonObjectError(input, options) {
  return applyOptionsToDecoderErrors(
    input,
    new DecoderError(input, 'invalid type', 'must be a non-null object', {
      decoderName: decoderName$3,
    }),
    options,
  );
}
function dictionaryD(decoder, optionalA, optionalB) {
  let keyDecoder;
  let options = {};
  if (optionalA) {
    if (optionalA instanceof Decoder || optionalA instanceof AsyncDecoder) {
      keyDecoder = optionalA;
    } else {
      options = optionalA;
    }
  }
  if (optionalB) {
    options = optionalB;
  }
  options = Object.assign({}, options, {
    decoderName: options.decoderName || decoderName$3,
  });
  if (decoder instanceof AsyncDecoder || keyDecoder instanceof AsyncDecoder) {
    if (options.allErrors) {
      return new AsyncDecoder(input =>
        __awaiter(this, void 0, void 0, function*() {
          if (typeof input !== 'object' || input === null) {
            return nonObjectError(input, options);
          }
          let hasError = false;
          const results = yield Promise.all(
            Object.entries(input).map(([entryKey, entryValue]) =>
              __awaiter(this, void 0, void 0, function*() {
                let key = entryKey;
                if (keyDecoder) {
                  const keyResult = yield asyncDecodeKey(
                    input,
                    keyDecoder,
                    entryKey,
                  );
                  if (Array.isArray(keyResult)) {
                    hasError = true;
                    return keyResult;
                  }
                  key = keyResult.value;
                }
                const valueResult = yield asyncDecodeValue(
                  input,
                  decoder,
                  entryValue,
                  entryKey,
                );
                if (Array.isArray(valueResult)) {
                  hasError = true;
                  return valueResult;
                }
                return [key, valueResult.value];
              }),
            ),
          );
          if (hasError) {
            const errors = [];
            results.forEach(result => {
              if (result[0] instanceof DecoderError) {
                errors.push(...result);
              }
            });
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          return ok(Object.fromEntries(results));
        }),
      );
    }
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        if (typeof input !== 'object' || input === null) {
          return nonObjectError(input, options);
        }
        const resultObject = {};
        for (const [entryKey, entryValue] of Object.entries(input)) {
          let key = entryKey;
          if (keyDecoder) {
            const keyResult = yield asyncDecodeKey(input, keyDecoder, entryKey);
            if (Array.isArray(keyResult)) {
              return applyOptionsToDecoderErrors(input, keyResult, options);
            }
            key = keyResult.value;
          }
          const valueResult = yield asyncDecodeValue(
            input,
            decoder,
            entryValue,
            entryKey,
          );
          if (Array.isArray(valueResult)) {
            return applyOptionsToDecoderErrors(input, valueResult, options);
          }
          resultObject[key] = valueResult.value;
        }
        return ok(resultObject);
      }),
    );
  }
  return new Decoder(input => {
    if (typeof input !== 'object' || input === null) {
      return nonObjectError(input, options);
    }
    const resultObject = {};
    const entries = Object.entries(input);
    const allErrors = [];
    for (const [entryKey, entryValue] of entries) {
      let key = entryKey;
      if (keyDecoder) {
        const keyResult = keyDecoder.decode(entryKey);
        if (areDecoderErrors(keyResult)) {
          const errors = keyResult.map(error =>
            childKeyError(input, error, entryKey),
          );
          if (options.allErrors) {
            allErrors.push(...errors);
            continue;
          }
          return applyOptionsToDecoderErrors(input, errors, options);
        }
        key = keyResult.value;
      }
      const valueResult = decoder.decode(entryValue);
      if (areDecoderErrors(valueResult)) {
        const errors = valueResult.map(error =>
          childValueError(input, error, entryKey),
        );
        if (options.allErrors) {
          allErrors.push(...errors);
          continue;
        }
        return applyOptionsToDecoderErrors(input, errors, options);
      }
      resultObject[key] = valueResult.value;
    }
    if (allErrors.length > 0) {
      return applyOptionsToDecoderErrors(input, allErrors, options);
    }
    return ok(resultObject);
  });
}

const decoderName$4 = 'matchD';
/**
 * Can be used to verify that an unknown value is a `string`
 * which conforms to the given `RegExp`.
 */
function matchD(regex, options = {}) {
  return new Decoder(value => {
    if (typeof value !== 'string') {
      return err(
        value,
        'invalid type',
        `must be a string`,
        decoderName$4,
        options,
      );
    }
    return regex.test(value)
      ? ok(value)
      : err(
          value,
          'invalid value',
          `must be a string matching the pattern "${regex}"`,
          decoderName$4,
          options,
        );
  });
}

// this email regex is taken from this S.O. answer
// https://dba.stackexchange.com/a/165923/191558
// which in turn takes it from the HTML5 spec
const emailRegex = /^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/;
/** Can be used to verify that an unknown value is an email address `string`. */
function emailD(options = {}) {
  return matchD(emailRegex, {
    data: options.data,
    decoderName: options.decoderName || 'emailD',
    errorMsg: options.errorMsg || 'must be a email address',
  });
}

function exactlyD(exact, options = {}) {
  return new Decoder(value =>
    value === exact
      ? ok(value)
      : err(value, 'invalid value', `must be ${exact}`, 'exactlyD', options),
  );
}

/**
 * Accepts a javascript constructor argument (`class`) and creates a
 * decoder which verifies that its input is an `instanceof` the
 * provided constructor.
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
function instanceOfD(clazz, options = {}) {
  return new Decoder(value =>
    value instanceof clazz
      ? ok(value)
      : err(
          value,
          'invalid class',
          `must be an instance of ${clazz.name}`,
          'instanceOfD',
          options,
        ),
  );
}

/**
 * Can be used to verify that an unknown value is a whole `number`.
 *
 * Related:
 * - numberD
 * */
function integerD(options = {}) {
  return new Decoder(value =>
    Number.isInteger(value)
      ? ok(value)
      : err(
          value,
          'invalid type',
          'must be a whole number',
          'integerD',
          options,
        ),
  );
}

function lazyD(decoderFn, options = {}) {
  if (options.promise) {
    return new AsyncDecoder(value =>
      __awaiter(this, void 0, void 0, function*() {
        const decoder = yield decoderFn(value);
        return decoder.decode(value);
      }),
    );
  }
  return new Decoder(value => decoderFn(value).decode(value));
}

const decoderName$5 = 'predicateD';
const defaultMsg = 'failed custom check';
function predicateD(fn, options = {}) {
  if (options.promise) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        return (yield fn(input))
          ? ok(input)
          : err(input, 'failed check', defaultMsg, decoderName$5, options);
      }),
    );
  }
  return new Decoder(input =>
    fn(input)
      ? ok(input)
      : err(input, 'failed check', defaultMsg, decoderName$5, options),
  );
}

const decoderName$6 = 'optionalD';
const undefinedDecoder = exactlyD(undefined);
const nullDecoder = exactlyD(null);
function optionalD(decoder, options = {}) {
  if (decoder instanceof AsyncDecoder) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        let result = undefinedDecoder.decode(input);
        if (isDecoderSuccess(result)) return result;
        result = nullDecoder.decode(input);
        if (isDecoderSuccess(result)) return result;
        result = yield decoder.decode(input);
        if (isDecoderSuccess(result)) return result;
        return applyOptionsToDecoderErrors(
          input,
          Array.from(result.values()).map(
            error =>
              new DecoderError(
                input,
                'invalid type',
                `${error.message} OR must be null OR must be undefined`,
                {
                  child: error,
                  decoderName: decoderName$6,
                },
              ),
          ),
          options,
        );
      }),
    );
  }
  return new Decoder(input => {
    let result = undefinedDecoder.decode(input);
    if (isDecoderSuccess(result)) return result;
    result = nullDecoder.decode(input);
    if (isDecoderSuccess(result)) return result;
    result = decoder.decode(input);
    if (isDecoderSuccess(result)) return result;
    return applyOptionsToDecoderErrors(
      input,
      Array.from(result.values()).map(
        error =>
          new DecoderError(
            input,
            'invalid type',
            `${error.message} OR must be null OR must be undefined`,
            {
              child: error,
              decoderName: decoderName$6,
            },
          ),
      ),
      options,
    );
  });
}

/**
 * Creates a decoder which always returns `DecoderError` with
 * whatever input value is provided to it.
 */
function neverD(options = {}) {
  return new Decoder(value =>
    err(value, 'forbidden', 'must not be present', 'neverD', options),
  );
}

const decoderName$7 = 'nullableD';
const nullDecoder$1 = exactlyD(null);
function nullableD(decoder, options = {}) {
  if (decoder instanceof AsyncDecoder) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        let result = nullDecoder$1.decode(input);
        if (isDecoderSuccess(result)) return result;
        result = yield decoder.decode(input);
        if (isDecoderSuccess(result)) return result;
        return applyOptionsToDecoderErrors(
          input,
          Array.from(result.values()).map(
            error =>
              new DecoderError(
                input,
                'invalid type',
                `${error.message} OR must be null`,
                {
                  child: error,
                  decoderName: decoderName$7,
                },
              ),
          ),
          options,
        );
      }),
    );
  }
  return new Decoder(input => {
    let result = nullDecoder$1.decode(input);
    if (isDecoderSuccess(result)) return result;
    result = decoder.decode(input);
    if (isDecoderSuccess(result)) return result;
    return applyOptionsToDecoderErrors(
      input,
      Array.from(result.values()).map(
        error =>
          new DecoderError(
            input,
            'invalid type',
            `${error.message} OR must be null`,
            {
              child: error,
              decoderName: decoderName$7,
            },
          ),
      ),
      options,
    );
  });
}

/**
 * Can be used to verify that an unknown value is a `number`.
 *
 * Related:
 * - integerD
 */
function numberD(options = {}) {
  return new Decoder(value =>
    Number.isFinite(value)
      ? ok(value)
      : err(value, 'invalid type', 'must be a number', 'numberD', options),
  );
}

const decoderName$8 = 'objectD';
function nonObjectError$1(input, options) {
  return applyOptionsToDecoderErrors(
    input,
    [
      new DecoderError(input, 'invalid type', 'must be a non-null object', {
        decoderName: decoderName$8,
      }),
    ],
    options,
  );
}
function unknownKeyError(value, key) {
  return new DecoderError(value, 'unknown key', `unknown key ["${key}"]`, {
    decoderName: decoderName$8,
    location: errorLocation(key, ''),
    key,
  });
}
function fromChildError(child, value, key) {
  let type;
  let message;
  let location;
  let errorKey = key;
  if (value.hasOwnProperty(key)) {
    location = errorLocation(key, child.location);
    type = 'invalid key value';
    message = `invalid value for key ["${key}"] > ${child.message}`;
  } else {
    type = 'missing key';
    message = `missing required key ["${key}"]`;
    location = undefined;
    errorKey = undefined;
  }
  return new DecoderError(value, type, message, {
    decoderName: decoderName$8,
    child,
    location,
    key: errorKey,
  });
}
function checkInputKeys(decoderObject, input, options) {
  const expectedkeys = Object.keys(decoderObject);
  const actualkeys = Object.keys(input);
  if (!options.allErrors) {
    const invalidKey = actualkeys.find(key => !expectedkeys.includes(key));
    if (invalidKey !== undefined) {
      return [unknownKeyError(input, invalidKey)];
    }
  } else {
    const invalidKeys = actualkeys.filter(key => !expectedkeys.includes(key));
    const errors = invalidKeys.map(key => unknownKeyError(input, key));
    if (errors.length > 0) return errors;
  }
}
function objectD(decoderObject, options = {}) {
  const hasAsyncDecoder = Object.values(decoderObject).some(
    decoder => decoder instanceof AsyncDecoder,
  );
  options = Object.assign({}, options, {
    decoderName: options.decoderName || decoderName$8,
  });
  if (hasAsyncDecoder) {
    if (options.allErrors) {
      return new AsyncDecoder(input =>
        __awaiter(this, void 0, void 0, function*() {
          if (typeof input !== 'object' || input === null) {
            return nonObjectError$1(input, options);
          }
          const allErrors = [];
          if (options.noExcessProperties) {
            const invalidKeys = checkInputKeys(decoderObject, input, options);
            if (invalidKeys) {
              allErrors.push(...invalidKeys);
            }
          }
          // eslint-disable-next-line @typescript-eslint/no-explicit-any
          const entries = Object.entries(decoderObject);
          let resolvedEntries = yield Promise.all(
            entries.map(
              // eslint-disable-next-line @typescript-eslint/no-explicit-any
              ([key, decoder]) =>
                __awaiter(this, void 0, void 0, function*() {
                  // eslint-disable-next-line @typescript-eslint/no-explicit-any
                  const value = input[key];
                  const result = yield decoder.decode(value);
                  if (areDecoderErrors(result)) {
                    const errors = result.map(error =>
                      fromChildError(error, input, key),
                    );
                    allErrors.push(...errors);
                    return;
                  }
                  // eslint-disable-next-line @typescript-eslint/no-explicit-any
                  return [key, result.value];
                }),
            ),
          );
          if (allErrors.length > 0) {
            return applyOptionsToDecoderErrors(input, allErrors, options);
          }
          if (options.removeUndefinedProperties) {
            resolvedEntries = resolvedEntries.filter(
              ([, value]) => value !== undefined,
            );
          }
          return ok(
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            Object.fromEntries(resolvedEntries),
          );
        }),
      );
    }
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        if (typeof input !== 'object' || input === null) {
          return nonObjectError$1(input, options);
        }
        if (options.noExcessProperties) {
          const invalidKeys = checkInputKeys(decoderObject, input, options);
          if (invalidKeys) {
            return applyOptionsToDecoderErrors(input, invalidKeys, options);
          }
        }
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const entries = Object.entries(decoderObject);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const resultObj = {};
        for (const [key, decoder] of entries) {
          // eslint-disable-next-line @typescript-eslint/no-explicit-any
          const value = input[key];
          const result = yield decoder.decode(value);
          if (areDecoderErrors(result)) {
            const errors = result.map(error =>
              fromChildError(error, input, key),
            );
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          if (options.removeUndefinedProperties && result.value === undefined) {
            continue;
          }
          resultObj[key] = result.value;
        }
        return ok(resultObj);
      }),
    );
  }
  return new Decoder(input => {
    if (typeof input !== 'object' || input === null) {
      return nonObjectError$1(input, options);
    }
    const allErrors = [];
    if (options.noExcessProperties) {
      const invalidKeys = checkInputKeys(decoderObject, input, options);
      if (invalidKeys) {
        if (!options.allErrors) {
          return applyOptionsToDecoderErrors(input, invalidKeys, options);
        }
        allErrors.push(...invalidKeys);
      }
    }
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const entries = Object.entries(decoderObject);
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const resultObj = {};
    for (const [key, decoder] of entries) {
      // eslint-disable-next-line @typescript-eslint/no-explicit-any
      const value = input[key];
      const result = decoder.decode(value);
      if (areDecoderErrors(result)) {
        const errors = result.map(error => fromChildError(error, input, key));
        if (options.allErrors) {
          allErrors.push(...errors);
          continue;
        }
        return applyOptionsToDecoderErrors(input, errors, options);
      }
      if (options.removeUndefinedProperties && result.value === undefined) {
        continue;
      }
      resultObj[key] = result.value;
    }
    if (allErrors.length > 0) {
      return applyOptionsToDecoderErrors(input, allErrors, options);
    }
    return ok(resultObj);
  });
}

const decoderName$9 = 'undefinableD';
const undefinedDecoder$1 = exactlyD(undefined);
const errorsMapFn = input => error =>
  new DecoderError(
    input,
    'invalid type',
    `${error.message} OR must be undefined`,
    {
      child: error,
      decoderName: decoderName$9,
    },
  );
function undefinableD(decoder, options = {}) {
  if (decoder instanceof AsyncDecoder) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        let result = undefinedDecoder$1.decode(input);
        if (isDecoderSuccess(result)) return result;
        result = yield decoder.decode(input);
        if (isDecoderSuccess(result)) return result;
        return applyOptionsToDecoderErrors(
          input,
          result.map(errorsMapFn(input)),
          options,
        );
      }),
    );
  }
  return new Decoder(input => {
    let result = undefinedDecoder$1.decode(input);
    if (isDecoderSuccess(result)) return result;
    result = decoder.decode(input);
    if (isDecoderSuccess(result)) return result;
    return applyOptionsToDecoderErrors(
      input,
      result.map(errorsMapFn(input)),
      options,
    );
  });
}

/** Can be used to verify that an unknown value is a `string`. */
function stringD(options = {}) {
  return new Decoder(value =>
    typeof value === 'string'
      ? ok(value)
      : err(value, 'invalid type', 'must be a string', 'stringD', options),
  );
}

const decoderName$a = 'tupleD';
function nonArrayError$1(input, options) {
  return err(input, 'invalid type', 'must be an array', decoderName$a, options);
}
function invalidLengthError(length, input) {
  return new DecoderError(
    input,
    'invalid length',
    `array must have length ${length}`,
    {
      decoderName: decoderName$a,
    },
  );
}
function childError$1(child, value, key) {
  const location = errorLocation(key, child.location);
  return new DecoderError(
    value,
    'invalid element',
    `invalid element [${key}] > ${child.message}`,
    {
      decoderName: decoderName$a,
      child,
      location,
      key,
    },
  );
}
function tupleD(decoders, options = {}) {
  const hasAsyncDecoder = decoders.some(
    decoder => decoder instanceof AsyncDecoder,
  );
  options = Object.assign({}, options, {
    decoderName: options.decoderName || decoderName$a,
  });
  if (hasAsyncDecoder) {
    if (options.allErrors) {
      return new AsyncDecoder(input =>
        __awaiter(this, void 0, void 0, function*() {
          if (!Array.isArray(input)) {
            return nonArrayError$1(input, options);
          }
          let hasError = false;
          const results = yield Promise.all(
            decoders.map((decoder, index) =>
              __awaiter(this, void 0, void 0, function*() {
                const result = yield decoder.decode(input[index]);
                if (areDecoderErrors(result)) {
                  hasError = true;
                  return result.map(error => childError$1(error, input, index));
                }
                return result;
              }),
            ),
          );
          if (hasError || input.length !== decoders.length) {
            const errors = results
              .filter(result => Array.isArray(result))
              .flat();
            if (input.length !== decoders.length) {
              errors.push(invalidLengthError(decoders.length, input));
            }
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          return ok(results.map(result => result.value));
        }),
      );
    }
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        if (!Array.isArray(input)) {
          return nonArrayError$1(input, options);
        } else if (input.length !== decoders.length) {
          return applyOptionsToDecoderErrors(
            input,
            invalidLengthError(decoders.length, input),
            options,
          );
        }
        const tuple = [];
        let index = -1;
        for (const decoder of decoders) {
          index++;
          const result = yield decoder.decode(input[index]);
          if (areDecoderErrors(result)) {
            const errors = result.map(error =>
              childError$1(error, input, index),
            );
            return applyOptionsToDecoderErrors(input, errors, options);
          }
          tuple.push(result.value);
        }
        return ok(tuple);
      }),
    );
  }
  return new Decoder(input => {
    if (!Array.isArray(input)) {
      return nonArrayError$1(input, options);
    }
    const tuple = [];
    const allErrors = [];
    let index = -1;
    if (input.length !== decoders.length) {
      if (!options.allErrors) {
        return applyOptionsToDecoderErrors(
          input,
          invalidLengthError(decoders.length, input),
          options,
        );
      }
      allErrors.push(invalidLengthError(decoders.length, input));
    }
    for (const decoder of decoders) {
      index++;
      const result = decoder.decode(input[index]);
      if (areDecoderErrors(result)) {
        const errors = result.map(error => childError$1(error, input, index));
        if (options.allErrors) {
          allErrors.push(...errors);
          continue;
        }
        return applyOptionsToDecoderErrors(input, errors, options);
      }
      tuple.push(result.value);
    }
    if (allErrors.length > 0) {
      return applyOptionsToDecoderErrors(input, allErrors, options);
    }
    return ok(tuple);
  });
}

/** Can be used to verify that an unknown value is a uuid v4 `string`. */
function uuidD(options = {}) {
  return matchD(
    /[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89aAbB][a-f0-9]{3}-[a-f0-9]{12}/i,
    {
      data: options.data,
      decoderName: options.decoderName || 'uuidD',
      errorMsg: options.errorMsg || 'must be a v4 UUID',
    },
  );
}

export {
  anyD,
  anyOfD,
  applyOptionsToDecoderErrors,
  arrayD,
  booleanD,
  chainOfD,
  constantD,
  dictionaryD,
  emailD,
  exactlyD,
  instanceOfD,
  integerD,
  lazyD,
  matchD,
  neverD,
  nullableD,
  numberD,
  objectD,
  optionalD,
  predicateD,
  raceToDecoderSuccess,
  stringD,
  tupleD,
  undefinableD,
  uuidD,
};
