import { Decoder, AsyncDecoder, DecoderError } from 'ts-decoders';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
  return new (P || (P = Promise))(function(resolve, reject) {
    function fulfilled(value) {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    }
    function rejected(value) {
      try {
        step(generator['throw'](value));
      } catch (e) {
        reject(e);
      }
    }
    function step(result) {
      result.done
        ? resolve(result.value)
        : new P(function(resolve) {
            resolve(result.value);
          }).then(fulfilled, rejected);
    }
    step((generator = generator.apply(thisArg, _arguments || [])).next());
  });
}

function printValue(value) {
  return value instanceof Promise ? value : JSON.stringify(value);
}
/**
 * Make an assertion that `actual` is a Decoder.
 *
 * If not then throw.
 */
expect.extend({
  toBeDecoder(received) {
    return {
      message: () => `"${printValue(received)}" expected to be a Decoder`,
      pass: received instanceof Decoder,
    };
  },
});
/**
 * Make an assertion that `actual` is a AsyncDecoder.
 *
 * If not then throw.
 */
expect.extend({
  toBeAsyncDecoder(received) {
    return {
      message: () => `"${printValue(received)}" expected to be a AsyncDecoder`,
      pass: received instanceof AsyncDecoder,
    };
  },
});
/**
 * Make an assertion that `actual` is a DecoderSuccess.
 *
 * If not then throw.
 */
expect.extend({
  toBeDecoderSuccess(received, expected) {
    const pass = !(received instanceof Promise);
    if (!pass) {
      return {
        message: () => `"${printValue(received)}" expected to not be a promise`,
        pass: false,
      };
    }
    expect(received).toEqual(expected);
    return {
      message: `"${printValue(received)}" expected to equal ${printValue(
        expected,
      )}`,
      pass: true,
    };
  },
});
/**
 * Make an assertion that `actual` is a DecoderSuccess
 * with matching `value` to `expected`.
 *
 * If not then throw.
 */
expect.extend({
  toBeAsyncDecoderSuccess(received, expected) {
    return __awaiter(this, void 0, void 0, function*() {
      const pass = received instanceof Promise;
      if (!pass) {
        return {
          message: () => `"${printValue(received)}" expected to be a promise`,
          pass: false,
        };
      }
      const value = yield received;
      expect(value).toEqual(expected);
      return {
        message: `"${printValue(value)}" expected to equal ${printValue(
          expected,
        )}`,
        pass: true,
      };
    });
  },
});
/**
 * Make an assertion that `actual` is an array of DecoderError
 * with matching properties and size to `expected`. Also asserts that
 * `actual` is NOT an instanceof `Promise`.
 *
 * If not then throw.
 */
expect.extend({
  toBeDecoderErrors(received, expected) {
    if (received instanceof Promise) {
      return {
        message: () => `"${printValue(received)}" expected to not be a promise`,
        pass: false,
      };
    }
    if (
      !Array.isArray(received) ||
      received.some(val => !(val instanceof DecoderError))
    ) {
      return {
        message: () =>
          `${printValue(received)} expected to be an array of DecoderError`,
        pass: false,
      };
    }
    const normExpected =
      expected instanceof DecoderError ? [expected] : expected;
    if (received.length !== normExpected.length) {
      return {
        message: () =>
          `${printValue(received)} expected to have length ${
            normExpected.length
          } but had length ${received.length}`,
        pass: false,
      };
    }
    normExpected.forEach(value => {
      expect(received).toContainEqual(value);
    });
    return {
      message: `${printValue(received)} expected to contain items`,
      pass: true,
    };
  },
});
/**
 * Similar to `assertDecoderErrors`, except also asserts that
 * `actual` is an instanceof `Promise`.
 */
expect.extend({
  toBeAsyncDecoderErrors(received, expected) {
    return __awaiter(this, void 0, void 0, function*() {
      if (!(received instanceof Promise)) {
        return {
          message: () => `"${printValue(received)}" expected to be a promise`,
          pass: false,
        };
      }
      const value = yield received;
      expect(value).toBeDecoderErrors(expected);
      return {
        message: `${printValue(
          value,
        )} expected to be an array of decoder errors`,
        pass: true,
      };
    });
  },
});
/**
 * Make an assertion that `decoder.decode(value)` resolves to a
 * DecoderSuccess with matching `value` to `expected`.
 */
expect.extend({
  toDecodeSuccessfully([decoder, value], expected) {
    expect(decoder.decode(value)).toBeDecoderSuccess(expected);
    return {
      message: `${printValue(value)} expected to decode successfully`,
      pass: true,
    };
  },
});
expect.extend({
  toAsyncDecodeSuccessfully([decoder, value], expected) {
    return __awaiter(this, void 0, void 0, function*() {
      if (!(decoder instanceof AsyncDecoder || value instanceof Promise)) {
        return {
          message: () =>
            `${decoder &&
              decoder.constructor.name} expected to be a AsyncDecoder`,
          pass: false,
        };
      }
      yield expect(decoder.decode(value)).toBeAsyncDecoderSuccess(expected);
      return {
        message: `${printValue(value)} expected to async decode successfully`,
        pass: true,
      };
    });
  },
});
/**
 * Make an assertion that `decoder.decode(value)` resolves to an
 * array of DecoderError with matching properties and size to
 * `expected`.
 *
 * If not then throw.
 */
expect.extend({
  toDecodeWithErrors([decoder, value], expected) {
    expect(decoder.decode(value)).toBeDecoderErrors(expected);
    return {
      message: `${printValue(value)} expected to decode with errors`,
      pass: true,
    };
  },
});
expect.extend({
  toAsyncDecodeWithErrors([decoder, value], expected) {
    return __awaiter(this, void 0, void 0, function*() {
      if (!(decoder instanceof AsyncDecoder || value instanceof Promise)) {
        return {
          message: () =>
            `${decoder.constructor.name} expected to be a AsyncDecoder`,
          pass: false,
        };
      }
      yield expect(decoder.decode(value)).toBeAsyncDecoderErrors(expected);
      return {
        message: `${printValue(value)} expected to async decode with errors`,
        pass: true,
      };
    });
  },
});
