/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
  return new (P || (P = Promise))(function(resolve, reject) {
    function fulfilled(value) {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    }
    function rejected(value) {
      try {
        step(generator['throw'](value));
      } catch (e) {
        reject(e);
      }
    }
    function step(result) {
      result.done
        ? resolve(result.value)
        : new P(function(resolve) {
            resolve(result.value);
          }).then(fulfilled, rejected);
    }
    step((generator = generator.apply(thisArg, _arguments || [])).next());
  });
}

/**
 * DecoderSuccess returned on a successful call to
 * `Decoder#decode` or `AsyncDecoder#decode`
 */
class DecoderSuccess {
  constructor(value) {
    this.value = value;
  }
}
/**
 * DecoderError returned on a failed call to
 * `Decoder#decode` or `AsyncDecoder#decode`
 */
class DecoderError {
  constructor(
    /** The input that failed validation. */
    input,
    /** The type of error. e.g. "invalid type" */
    type,
    /** A human readable error message. e.g. "must be a string" */
    message,
    options = {},
  ) {
    this.input = input;
    this.type = type;
    this.message = message;
    this.decoderName = options.decoderName || '';
    this.location = options.location || '';
    this.child = options.child;
    this.key = options.key;
    this.data = options.data;
  }
  /**
   * Starting with this error, an array of the keys associated with
   * this error as well as all child errors.
   */
  path() {
    if (this.key === undefined) {
      if (!this.child) return [];
      return this.child.path();
    }
    if (!this.child) return [this.key];
    return [this.key, ...this.child.path()];
  }
}
function isDecoderSuccess(result) {
  return result instanceof DecoderSuccess;
}
function areDecoderErrors(result) {
  return Array.isArray(result);
}

function coerceDecodeFnResult(result) {
  return result instanceof DecoderError ? [result] : result;
}
/**
 * An object which can be used to validate and process `unknown` values
 * and cast them to the appropriate typescript type.
 *
 * @param decodeFn the function which is used to decode input values
 */
class Decoder {
  constructor(decodeFn) {
    this.decodeFn = decodeFn;
  }
  decode(input) {
    if (input instanceof Promise) {
      return input.then(res => coerceDecodeFnResult(this.decodeFn(res)));
    }
    return coerceDecodeFnResult(this.decodeFn(input));
  }
  /**
   * On decode failure, handle the DecoderErrors.
   */
  catch(fn) {
    return new Decoder(input => {
      const result = coerceDecodeFnResult(this.decodeFn(input));
      if (areDecoderErrors(result)) {
        return fn(input, result);
      }
      return result;
    });
  }
  /**
   * On decode success, transform a value using the provided
   * transformation function.
   */
  map(fn) {
    return new Decoder(input => {
      const result = this.decodeFn(input);
      if (result instanceof DecoderSuccess) {
        return new DecoderSuccess(fn(result.value));
      }
      return result;
    });
  }
  chain(a) {
    const fn = a instanceof Decoder ? a.decodeFn : a;
    return new Decoder(input => {
      const result = this.decodeFn(input);
      if (result instanceof DecoderSuccess) {
        return fn(result.value);
      }
      return result;
    });
  }
  toAsyncDecoder() {
    const fn = this.decodeFn;
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        return fn(input);
      }),
    );
  }
}
/**
 * An object which can be used to validate and process `unknown` values
 * and cast them to the appropriate typescript type. Unlike `Decoder`,
 * `AsyncDecoder` can receive a `decodeFn` which returns a promise.
 *
 * @param decodeFn the function which is used to decode input values
 */
class AsyncDecoder {
  constructor(decodeFn) {
    this.decodeFn = decodeFn;
  }
  decode(input) {
    return __awaiter(this, void 0, void 0, function*() {
      return coerceDecodeFnResult(yield this.decodeFn(yield input));
    });
  }
  /**
   * On decode failure, handle the DecoderErrors.
   */
  catch(fn) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        const result = coerceDecodeFnResult(yield this.decodeFn(input));
        if (areDecoderErrors(result)) {
          return fn(input, result);
        }
        return result;
      }),
    );
  }
  /**
   * On decode success, transform a value using the provided
   * transformation function.
   */
  map(fn) {
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        const result = yield this.decodeFn(input);
        if (result instanceof DecoderSuccess) {
          return new DecoderSuccess(yield fn(result.value));
        }
        return result;
      }),
    );
  }
  chain(a) {
    const fn =
      a instanceof Decoder || a instanceof AsyncDecoder ? a.decodeFn : a;
    return new AsyncDecoder(input =>
      __awaiter(this, void 0, void 0, function*() {
        const result = yield this.decodeFn(input);
        if (result instanceof DecoderSuccess) {
          return fn(result.value);
        }
        return result;
      }),
    );
  }
}

class AssertDecoderError extends Error {
  constructor(errors) {
    super();
    this.errors = errors;
    this.name = 'AssertDecoderError';
    this.message = errors.map(error => error.message).join(' \n');
  }
}
function handleResult(result) {
  if (areDecoderErrors(result)) throw new AssertDecoderError(result);
  return result.value;
}
function assert(decoder) {
  if (decoder instanceof AsyncDecoder) {
    return value =>
      __awaiter(this, void 0, void 0, function*() {
        return handleResult(yield decoder.decode(yield value));
      });
  }
  const decode = value => handleResult(decoder.decode(value));
  // I have no idea how to explicitly type this return value
  // such that typescript does not complain
  // eslint-disable-next-line @typescript-eslint/no-explicit-any
  return input =>
    input instanceof Promise
      ? input.then(value => decode(value))
      : decode(input);
}

export {
  AssertDecoderError,
  AsyncDecoder,
  Decoder,
  DecoderError,
  DecoderSuccess,
  areDecoderErrors,
  assert,
  isDecoderSuccess,
};
