import {
  Decoder,
  DecoderSuccess,
  DecoderError,
  areDecoderErrors,
  AsyncDecoder,
} from 'ts-decoders';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
  return new (P || (P = Promise))(function(resolve, reject) {
    function fulfilled(value) {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    }
    function rejected(value) {
      try {
        step(generator['throw'](value));
      } catch (e) {
        reject(e);
      }
    }
    function step(result) {
      result.done
        ? resolve(result.value)
        : new P(function(resolve) {
            resolve(result.value);
          }).then(fulfilled, rejected);
    }
    step((generator = generator.apply(thisArg, _arguments || [])).next());
  });
}

/**
 * Test Decoders
 */
/** Simple boolean Decoder for use in tests */
const BOOLEAN_DEC = new Decoder(value =>
  typeof value === 'boolean'
    ? new DecoderSuccess(value)
    : new DecoderError(value, 'invalid type', 'must be a boolean'),
);
/** Simple string Decoder for use in tests */
const STRING_DEC = new Decoder(value =>
  typeof value === 'string'
    ? new DecoderSuccess(value)
    : new DecoderError(value, 'invalid type', 'must be a string'),
);
/** Simple number Decoder for use in tests */
const NUMBER_DEC = new Decoder(value =>
  typeof value === 'number'
    ? new DecoderSuccess(value)
    : new DecoderError(value, 'invalid type', 'must be a number'),
);
/** Simple `any` Decoder for use in tests */
const ANY_DEC = new Decoder(value => new DecoderSuccess(value));
/** Recursive "nested arrays" Decoder for use in tests */
const RECURSIVE_DEC = new Decoder(input => {
  if (!Array.isArray(input)) {
    return new DecoderError(input, 'invalid type', 'must be an array');
  }
  let index = -1;
  const array = [];
  for (const element of input) {
    index++;
    const result = RECURSIVE_DEC.decode(element);
    if (areDecoderErrors(result)) {
      return result.map(
        error =>
          new DecoderError(
            element,
            'invalid element',
            `invalid element [${index}] > ${error.message}`,
            {
              location: `[${index}]${error.location}`,
              key: index,
              child: error,
            },
          ),
      );
    }
    array.push(result.value);
  }
  return new DecoderSuccess(array);
});
/**
 * Test AsyncDecoders
 */
/** Simple boolean AsyncDecoder for use in tests */
const BOOLEAN_ASYNC_DEC = new AsyncDecoder(value =>
  __awaiter(undefined, void 0, void 0, function*() {
    return typeof value === 'boolean'
      ? new DecoderSuccess(value)
      : new DecoderError(value, 'invalid type', 'must be a boolean');
  }),
);
/** Simple number AsyncDecoder for use in tests */
const STRING_ASYNC_DEC = new AsyncDecoder(value =>
  __awaiter(undefined, void 0, void 0, function*() {
    return typeof value === 'string'
      ? new DecoderSuccess(value)
      : new DecoderError(value, 'invalid type', 'must be a string');
  }),
);
/** Simple number AsyncDecoder for use in tests */
const NUMBER_ASYNC_DEC = new AsyncDecoder(value =>
  __awaiter(undefined, void 0, void 0, function*() {
    return typeof value === 'number'
      ? new DecoderSuccess(value)
      : new DecoderError(value, 'invalid type', 'must be a number');
  }),
);
/** Simple number AsyncDecoder for use in tests */
const ANY_ASYNC_DEC = new AsyncDecoder(value =>
  __awaiter(undefined, void 0, void 0, function*() {
    return new DecoderSuccess(value);
  }),
);
/** Recursive "nested array" AsyncDecoder for use in tests */
const RECURSIVE_ASYNC_DEC = new AsyncDecoder(input =>
  __awaiter(undefined, void 0, void 0, function*() {
    if (!Array.isArray(input)) {
      return new DecoderError(input, 'invalid type', 'must be an array');
    }
    let index = -1;
    const array = [];
    for (const element of input) {
      index++;
      const result = yield RECURSIVE_ASYNC_DEC.decode(element);
      if (areDecoderErrors(result)) {
        return result.map(
          error =>
            new DecoderError(
              element,
              'invalid element',
              `invalid element [${index}] > ${error.message}`,
              {
                location: `[${index}]${error.location}`,
                key: index,
                child: error,
              },
            ),
        );
      }
      array.push(result.value);
    }
    return new DecoderSuccess(array);
  }),
);

export {
  ANY_ASYNC_DEC,
  ANY_DEC,
  BOOLEAN_ASYNC_DEC,
  BOOLEAN_DEC,
  NUMBER_ASYNC_DEC,
  NUMBER_DEC,
  RECURSIVE_ASYNC_DEC,
  RECURSIVE_DEC,
  STRING_ASYNC_DEC,
  STRING_DEC,
};
